/*
   This is an implementation of the IntBag abstract data type (ADT).
   This class provides all the details of HOW an IntBag does what
   it does. In particular, this class specifies how the actual data
   will be stored (as a linked list).
*/

/**
* An IntLinkedBag is a collection of int numbers.
* The same number may appear multiple times in a bag.
*
* @note
*   (1) Beyond Int.MAX_VALUE elements, countOccurrences,
*   size, and grab are wrong.
*   <p>
*   (2) Because of the slow linear algorithms of this class, large bags will have
*   poor performance.
*
*/
public class IntLinkedBag extends IntBag
{
   // Invariant of the IntLinkedBag class:
   //   1. The elements in the bag are stored on a linked list.
   //   2. The head reference of the list is in the instance variable
   //      head.
   //   3. The total number of elements in the list is in the instance
   //      variable manyNodes.
   private IntNode head;
   private int manyNodes;


   /**
   * Initialize an empty bag.
   * @param - none
   * @postcondition
   *   This bag is empty.
   **/
   public IntLinkedBag( )
   {
      head = null;
      manyNodes = 0;
   }


   /**
   * Generate a copy of the given bag (as a IntLinkedBag).
   * @param  original
   *     IntBag to be copied. Subsequent changes to the copy
   *     will not affect this original, nor vice versa.
   **/
   public IntLinkedBag(IntBag original)
   {
      this.head = null;
      this.manyNodes = 0;
      this.add( original );
   }


   /**
   * {@inheritDoc}
   */
   public void add(int... elements)
   {
      for (int i : elements)
      {
         head = new IntNode(i, head);
         manyNodes++;
      }
   }


   /**
   * {@inheritDoc}
   */
   public void add(IntBag addend)
   {
      add( addend.toArray() );
   }


   /**
   * {@inheritDoc}
   */
   public int countOccurrences(int target)
   {
      int answer = 0;
      IntNode cursor = head;
      while (cursor != null)
      {
         if ( target == cursor.getData() )
         {
            answer++;
         }
         cursor = cursor.getLink( );
      }
      return answer;
   }


   /**
   * {@inheritDoc}
   */
   public boolean remove(int target)
   {
      boolean answer = false;
      for (IntNode cursor = head; cursor != null; cursor = cursor.getLink())
      {
         if ( target == cursor.getData() )
         {
            cursor.setData( head.getData() ); // copy the head node's data into this node
            head = head.getLink();            // remove the head node
            manyNodes--;
            answer = true;
            break;
         }
      }
      return answer;
   }


   /**
   * {@inheritDoc}
   */
   public int size( )
   {
      return manyNodes;
   }


   /**
   * {@inheritDoc}
   */
   public IntBag union(IntBag b2)
   {
      // use the copy constructor
      IntLinkedBag answer = new IntLinkedBag(this);

      // use add(IntBag)
      answer.add(b2);

      return answer;
   }


   /**
   * {@inheritDoc}
   */
   public int[] toArray()
   {
      int[] result = new int[manyNodes];
      IntNode cursor = head;
      for (int i = 0; i < manyNodes; i++)
      {
         result[i] = cursor.getData();
         cursor = cursor.getLink();
      }
      return result;
   }


   /**
   * {@inheritDoc}
   */
   public String toString()
   {
      String result = "[";
      IntNode cursor = head;
      for (int i = 0; i < manyNodes; i++)
      {
         if (i > 0)
         {
            result += " ";
         }
         result += cursor.getData();
         cursor = cursor.getLink();
         if (i < manyNodes-1)
         {
            result += ",";
         }
      }
      result += "]";
      return result;
   }
}
